/*
 * Copyright 2007-2010 Boris Kochergin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>
#include <vector>

#include <sys/socket.h>

#include <arpa/inet.h>
#ifdef __FreeBSD__
  #include <netinet/if_ether.h>
#endif
#ifdef __linux__
  #include <netinet/ether.h>
#endif

#include <stdint.h>

/*
 * Given a text representation of an Ethernet address in the form
 * "de:ad:be:ef:ca:fe", returns it in binary form.
 */
std::string textToEthernet(const std::string &textEthernetAddress) {
  ether_addr binaryEthernetAddress;
  ether_aton_r(textEthernetAddress.c_str(), &binaryEthernetAddress);
  #ifdef __FreeBSD__
    return std::string((char*)binaryEthernetAddress.octet, ETHER_ADDR_LEN);
  #endif
  #ifdef __linux__
    return std::string((char*)binaryEthernetAddress.ether_addr_octet,
                       ETHER_ADDR_LEN);
  #endif
}

/*
 * Given a binary representation of an Ethernet address, returns it in text
 * form.
 */
std::string ethernetToText(const char *binaryEthernetAddress) {
  ether_addr _binaryEthernetAddress;
  char textEthernetAddress[17];
  #ifdef __FreeBSD__
    memcpy(_binaryEthernetAddress.octet, binaryEthernetAddress, ETHER_ADDR_LEN);
  #endif
  #ifdef __linux__
    memcpy(_binaryEthernetAddress.ether_addr_octet, binaryEthernetAddress,
           ETHER_ADDR_LEN);
  #endif
  ether_ntoa_r(&_binaryEthernetAddress, textEthernetAddress);
  return textEthernetAddress;
}

/*
 * Given a text representation of an IP address in the form "192.168.0.1",
 * returns it in binary form, in network byte order.
 */
uint32_t textToIP(const std::string &textIP) {
  in_addr binaryIP;
  inet_pton(AF_INET, textIP.c_str(), &binaryIP);
  return binaryIP.s_addr;
}

/*
 * Given a binary representation of an IP address in network byte order, returns
 * it in text form.
 */
std::string ipToText(const uint32_t &binaryIP) {
  in_addr _binaryIP;
  char textIP[INET_ADDRSTRLEN];
  _binaryIP.s_addr = binaryIP;
  inet_ntop(AF_INET, &_binaryIP.s_addr, textIP, INET_ADDRSTRLEN);
  return textIP;
}
