/*
 * Copyright 2007-2010 Boris Kochergin. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string>
#include <vector>
#include <tr1/unordered_map>

#include <sys/socket.h>
#include <sys/time.h>

#include <arpa/inet.h>
#ifdef __FreeBSD__
  #include <netinet/if_ether.h>
#endif
#ifdef __linux__
  #include <netinet/ether.h>
#endif

#include <dnet.h>
#include <pcap.h>
#include <pthread.h>
#include <stdint.h>

enum ModeType { PASSIVE_MODE, AGGRESSIVE_MODE };
enum AttackType { GRATUITOUS_ARP_REQUEST_ATTACK, GRATUITOUS_ARP_REPLY_ATTACK,
                  ARP_REPLY_ATTACK };
enum GratuitousARPType { ARP_REQUEST, ARP_REPLY };

struct Interface {
  size_t index;
  pthread_t thread;
  std::string name;
  pcap_t *descriptor;
  std::string replyName;
  pcap_t *replyDescriptor;
  ModeType mode;
  std::tr1::unordered_map <uint32_t, std::string> pairs;
};

struct ARPCorrection {
  Interface *interface;
  /*
   * While, from a programming perspective, it might seem more convenient to
   * use an std::string for "sha", we may be storing a lot of these structures
   * in memory, so using a character array both lowers our memory requirements
   * and allows us to actually measure how much memory an instance of this
   * structure will use.
   */
  char sha[6];
  uint32_t spa;
  timeval attackTime;
};

/* Returns a pointer to the source address of an Ethernet frame. */
const char *getEthernetSourceAddress(const u_char *packet) {
  return (const char*)(packet + ETHER_ADDR_LEN);
}

/* Returns a pointer to the destination address of an Ethernet frame. */
const char *getEthernetDestinationAddress(const u_char *packet) {
  return (const char*)packet;
}

/*
 * Returns a pointer to the SHA (sender hardware address) field of an ARP
 * packet.
 */
const u_char *getARPSHA(const u_char *packet) {
  return packet + 22;
}

/*
 * Returns a pointer to the SPA (sender protocol address) field of an ARP
 * packet.
 */
const uint32_t *getARPSPA(const u_char *packet) {
  return (const uint32_t*)(packet + 28);
}

/*
 * Returns a pointer to the THA (target hardware address) field of an ARP
 * packet.
 */
const u_char* getARPTHA(const u_char *packet) {
  return packet + 32;
}

/*
 * Returns a pointer to the TPA (target protocol address) field of an ARP
 * packet.
 */
const uint32_t *getARPTPA(const u_char *packet) {
  return (const uint32_t*)(packet + 38);
}

/*
 * Returns the Ethernet address of the "interfaceName" interface upon success,
 * or "" upon failure.
 */
std::string getEthernetAddress(const std::string interfaceName) {
  intf_t *intf;
  intf_entry interface;
  intf = intf_open();
  if (intf == NULL) {
    return "";
  }
  interface.intf_len = sizeof(interface);
  strncpy(interface.intf_name, interfaceName.c_str(), INTF_NAME_LEN);
  if (intf_get(intf, &interface) < 0) {
    return "";
  }
  intf_close(intf);
  return std::string((const char*)interface.intf_link_addr.addr_eth.data,
                     ETHER_ADDR_LEN);
}

/*
 * Sends a gratuitous ARP packet using "pcapDescriptor", with an Ethernet
 * source address of "ethernetAddress", an Ethernet destination address of
 * ff:ff:ff:ff:ff:ff, an ARP sender hardware address of "sha", and an ARP
 * sender protocol address of "spa". If "type" is REQUEST, the ARP target
 * hardware address is set to 00:00:00:00:00:00 and an ARP target protocol
 * address is set to "spa". If "type" is REPLY, the ARP target hardware address
 * is set to ff:ff:ff:ff:ff:ff and the ARP target protocol address is set to
 * 0.0.0.0. Returns true upon success or false upon failure, in which case the
 * pcap_geterr() function may be used to retrieve the error message.
 */
bool sendGratuitousARPPacket(pcap_t *pcapDescriptor,
                             const std::string &ethernetAddress,
                             const char *sha, const uint32_t &spa,
                             const GratuitousARPType type) {
  /* Ethernet header. */
  /* Destination Ethernet address. */
  std::string packet("\xff\xff\xff\xff\xff\xff");
  /* Source Ethernet address. */
  packet.append(ethernetAddress);
  /* ARP frame type. */
  packet.append("\x08\x06", 2);
  /* ARP packet. */
  /* Hardware type (Ethernet). */
  packet.append("\x00\x01", 2);
  /* Protocol type (IP). */
  packet.append("\x08\x00", 2);
  /* Hardware size (6). */
  packet.append("\x06", 1);
  /* Protocol size (4). */
  packet.append("\x04", 1);
  /* Opcode. */
  switch (type) {
    /* Request. */
    case ARP_REQUEST:
      packet.append("\x00\x01", 2);
      break;
    /* Reply. */
    case ARP_REPLY:
      packet.append("\x00\x02", 2);
      break;
  }
  /* Sender hardware address. */
  packet.append(sha, ETHER_ADDR_LEN);
  /* Sender protocol address. */
  packet.append((const char*)&spa, 4);
  switch (type) {
    case ARP_REQUEST:
      /* Target hardware address. */
      packet.append("\x00\x00\x00\x00\x00\x00", ETHER_ADDR_LEN);
      /* Target protocol address. */
      packet.append((const char*)&spa, 4);
      break;
    case ARP_REPLY:
      /* Target hardware address. */
      packet.append("\xff\xff\xff\xff\xff\xff", ETHER_ADDR_LEN);
      /* Target protocol address. */
      packet.append("\x00\x00\x00\x00", 4);
      break;
  }
  if (pcap_inject(pcapDescriptor, (const void*)packet.c_str(),
                  packet.length()) == -1) {
    return false;
  }
  return true;
}
